@@ -19,6 +19,7 @@
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
+//#define ICECC_DEBUG 1
 #ifndef _GNU_SOURCE
 // getopt_long
 #define _GNU_SOURCE 1
@@ -34,6 +35,9 @@
 #include <netdb.h>
 #include <getopt.h>
 
+#ifdef HAVE_SIGNAL_H
+#include <signal.h>
+#endif
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
@@ -63,16 +67,15 @@
 #endif
 #include <netdb.h>
 
-#ifdef __FreeBSD__
-#  include <signal.h> // for kill(2)
-#  include <sys/time.h>
+#ifdef HAVE_SYS_RESOURCE_H
 #  include <sys/resource.h>
-#  ifndef RUSAGE_SELF
-#    define RUSAGE_SELF (0)
-#  endif
-#  ifndef RUSAGE_CHILDREN
-#    define RUSAGE_CHILDREN (-1)
-#  endif
+#endif
+
+#ifndef RUSAGE_SELF
+#  define RUSAGE_SELF (0)
+#endif
+#ifndef RUSAGE_CHILDREN
+#  define RUSAGE_CHILDREN (-1)
 #endif
 
 #include <deque>
@@ -80,6 +83,8 @@
 #include <algorithm>
 #include <ext/hash_set>
 #include <set>
+#include <fstream>
+#include <string>
 
 #include "ncpus.h"
 #include "exitcode.h"
@@ -89,8 +94,10 @@
 #include <comm.h>
 #include "load.h"
 #include "environment.h"
+#include "platform.h"
 
 const int PORT = 10245;
+static std::string pidFilePath;
 
 #ifndef __attribute_warn_unused_result__
 #define __attribute_warn_unused_result__
@@ -109,14 +116,15 @@ public:
      * JOBDONE: This was compiled by a local client and we got a jobdone - awaiting END
      * LINKJOB: This is a local job (aka link job) by a local client we told the scheduler about
      *          and await the finish of it
+     * TOINSTALL: We're receiving an environment transfer and wait for it to complete.
      * TOCOMPILE: We're supposed to compile it ourselves
      * WAITFORCS: Client asked for a CS and we asked the scheduler - waiting for its answer
      * WAITCOMPILE: Client got a CS and will ask him now (it's not me)
      * CLIENTWORK: Client is busy working and we reserve the spot (job_id is set if it's a scheduler job)
      * WAITFORCHILD: Client is waiting for the compile job to finish.
      */
-    enum Status { UNKNOWN, GOTNATIVE, PENDING_USE_CS, JOBDONE, LINKJOB, TOCOMPILE, WAITFORCS,
-                  WAITCOMPILE, CLIENTWORK, WAITFORCHILD, LASTSTATE=WAITFORCHILD } status;
+    enum Status { UNKNOWN, GOTNATIVE, PENDING_USE_CS, JOBDONE, LINKJOB, TOINSTALL, TOCOMPILE,
+                  WAITFORCS, WAITCOMPILE, CLIENTWORK, WAITFORCHILD, LASTSTATE=WAITFORCHILD } status;
     Client()
     {
         job_id = 0;
@@ -142,6 +150,8 @@ public:
             return "jobdone";
         case LINKJOB:
             return "linkjob";
+        case TOINSTALL:
+            return "toinstall";
         case TOCOMPILE:
             return "tocompile";
         case WAITFORCS:
@@ -168,14 +178,15 @@ public:
         job = 0;
 	if (pipe_to_child >= 0)
 	    close (pipe_to_child);
+
     }
     uint32_t job_id;
-    string outfile; // only useful for LINKJOB
+    string outfile; // only useful for LINKJOB or TOINSTALL
     MsgChannel *channel;
     UseCSMsg *usecsmsg;
     CompileJob *job;
     int client_id;
-    int pipe_to_child; // pipe to child process, only valid if WAITFORCHILD
+    int pipe_to_child; // pipe to child process, only valid if WAITFORCHILD or TOINSTALL
     pid_t child_pid;
 
     string dump() const
@@ -183,9 +194,11 @@ public:
         string ret = status_str( status ) + " " + channel->dump();
         switch ( status ) {
         case LINKJOB:
+            return ret + " CID: " + toString( client_id ) + " " + outfile;
+        case TOINSTALL:
             return ret + " " + toString( client_id ) + " " + outfile;
         case WAITFORCHILD:
-            return ret + " " + toString( client_id ) + " PID: " + toString( child_pid ) + " PFD: " + toString( pipe_to_child );
+            return ret + " CID: " + toString( client_id ) + " PID: " + toString( child_pid ) + " PFD: " + toString( pipe_to_child );
         default:
             if ( job_id ) {
                 string jobs;
@@ -338,6 +351,9 @@ static void dcc_daemon_terminate(int whichsig)
     if (am_parent) {
         /* kill whole group */
         kill(0, whichsig);
+
+        /* Remove pid file */
+        unlink(pidFilePath.c_str());
     }
 
     raise(whichsig);
@@ -348,7 +364,7 @@ void usage(const char* reason = 0)
   if (reason)
      cerr << reason << endl;
 
-  cerr << "usage: iceccd [-n <netname>] [-m <max_processes>] [-w] [-d|--daemonize] [-l logfile] [-s <schedulerhost>] [-v[v[v]]] [-r|--run-as-user] [-b <env-basedir>] [-u|--nobody-uid <nobody_uid>] [--cache-limit <MB>] [-N <node_name>]" << endl;
+  cerr << "usage: iceccd [-n <netname>] [-m <max_processes>] [--no-remote] [-w] [-d|--daemonize] [-l logfile] [-s <schedulerhost>] [-v[v[v]]] [-r|--run-as-user] [-b <env-basedir>] [-u|--nobody-uid <nobody_uid>] [--cache-limit <MB>] [-N <node_name>]" << endl;
   exit(1);
 }
 
@@ -396,7 +412,6 @@ int setup_listen_fd()
 
 
 struct timeval last_stat;
-time_t last_scheduler_ping;
 int mem_limit = 100;
 unsigned int max_kids = 0;
 
@@ -413,6 +428,7 @@ struct Daemon
     int listen_fd;
     string machine_name;
     string nodename;
+    bool noremote;
     bool custom_nodename;
     size_t cache_size;
     map<int, MsgChannel *> fd2chan;
@@ -441,6 +457,7 @@ struct Daemon
         new_client_id = 0;
         next_check = 0;
         cache_size = 0;
+        noremote = false;
         custom_nodename = false;
         icecream_load = 0;
         icecream_usage.tv_sec = icecream_usage.tv_usec = 0;
@@ -457,10 +474,12 @@ struct Daemon
     bool reannounce_environments() __attribute_warn_unused_result__;
     int answer_client_requests();
     bool handle_transfer_env( MsgChannel *c, Msg *msg ) __attribute_warn_unused_result__;
+    bool handle_transfer_env_done( Client *client ) __attribute_warn_unused_result__;
     bool handle_get_native_env( MsgChannel *c ) __attribute_warn_unused_result__;
     void handle_old_request();
     bool handle_compile_file( MsgChannel *c, Msg *msg ) __attribute_warn_unused_result__;
     bool handle_activity( MsgChannel *c ) __attribute_warn_unused_result__;
+    bool handle_file_chunk_env(MsgChannel* c, Msg *msg) __attribute_warn_unused_result__;
     void handle_end( Client *client, int exitcode );
     int scheduler_get_internals( ) __attribute_warn_unused_result__;
     void clear_children();
@@ -471,6 +490,8 @@ struct Daemon
     bool handle_compile_done (Client* client) __attribute_warn_unused_result__;
     int handle_cs_conf( ConfCSMsg *msg);
     string dump_internals() const;
+    string determine_nodename();
+    void determine_system();
     bool maybe_stats(bool force = false);
     bool send_scheduler(const Msg& msg) __attribute_warn_unused_result__;
     void close_scheduler();
@@ -478,6 +499,36 @@ struct Daemon
     int working_loop();
 };
 
+void Daemon::determine_system()
+{
+    struct utsname uname_buf;
+    if ( uname( &uname_buf ) ) {
+        log_perror( "uname call failed" );
+        return;
+    }
+
+    if ( nodename.length() && nodename != uname_buf.nodename )
+        custom_nodename  = true;
+
+    if (!custom_nodename)
+        nodename = uname_buf.nodename;
+
+    machine_name = determine_platform();
+}
+
+string Daemon::determine_nodename()
+{
+    if (custom_nodename && !nodename.empty())
+        return nodename;
+
+    // perhaps our host name changed due to network change?
+    struct utsname uname_buf;
+    if ( !uname( &uname_buf ) )
+        nodename = uname_buf.nodename;
+
+    return nodename;
+}
+
 bool Daemon::send_scheduler(const Msg& msg)
 {
     if (!scheduler) {
@@ -496,6 +547,7 @@ bool Daemon::send_scheduler(const Msg& msg)
 
 bool Daemon::reannounce_environments()
 {
+    log_error() << "reannounce_environments " << endl;
     LoginMsg lmsg( 0, nodename, "");
     lmsg.envs = available_environmnents(envbasedir);
     return send_scheduler( lmsg );
@@ -510,7 +562,6 @@ void Daemon::close_scheduler()
     scheduler = 0;
     delete discover;
     discover = 0;
-    clear_children();
 }
 
 bool Daemon::maybe_stats(bool send_ping)
@@ -518,14 +569,6 @@ bool Daemon::maybe_stats(bool send_ping)
     struct timeval now;
     gettimeofday( &now, 0 );
 
-    /* the scheduler didn't ping us for a long time, assume dead connection and recover */
-    if (!send_ping && now.tv_sec - last_scheduler_ping >= max_scheduler_ping + 2 * max_scheduler_pong) {
-        log_error() << "scheduler timeout.. " << now.tv_sec - last_scheduler_ping << " bigger than " <<
-            max_scheduler_ping + 2*max_scheduler_pong << " nuking" << endl;
-        close_scheduler();
-        return false;
-    }
-
     time_t diff_sent = ( now.tv_sec - last_stat.tv_sec ) * 1000 + ( now.tv_usec - last_stat.tv_usec ) / 1000;
     if ( diff_sent >= max_scheduler_pong * 1000 ) {
         StatsMsg msg;
@@ -533,7 +576,7 @@ bool Daemon::maybe_stats(bool send_ping)
         unsigned long idleLoad = 0;
         unsigned long niceLoad = 0;
 
-        if ( !fill_stats( idleLoad, niceLoad, memory_fillgrade, &msg ) )
+        if ( !fill_stats( idleLoad, niceLoad, memory_fillgrade, &msg, clients.active_processes ) )
             return false;
 
         time_t diff_stat = ( now.tv_sec - last_stat.tv_sec ) * 1000 + ( now.tv_usec - last_stat.tv_usec ) / 1000;
@@ -581,7 +624,7 @@ bool Daemon::maybe_stats(bool send_ping)
 #endif
 
         // Matz got in the urine that not all CPUs are always feed
-        mem_limit = std::max( msg.freeMem / std::min( std::max( max_kids, 1U ), 4U ), 100U );
+        mem_limit = std::max( int( msg.freeMem / std::min( std::max( max_kids, 1U ), 4U ) ), int( 100U ) );
 
         if ( abs(int(msg.load)-current_load) >= 100 || send_ping ) {
             if (!send_scheduler( msg ) )
@@ -591,13 +634,6 @@ bool Daemon::maybe_stats(bool send_ping)
         current_load = msg.load;
     }
 
-    if ( send_ping ) {
-        if (!send_scheduler(PingMsg()))
-            return false;
-
-        last_scheduler_ping = now.tv_sec;
-    }
-
     return true;
 }
 
@@ -628,7 +664,8 @@ string Daemon::dump_internals() const
         result += "  envs_last_use[" + it->first  + "] = " +
                   toString( it->second ) + "\n";
     }
-    result += "  Current kids: " + toString( current_kids ) + "\n";
+
+    result += "  Current kids: " + toString( current_kids ) + " (max: " + toString( max_kids ) + ")\n";
     if ( scheduler )
         result += "  Scheduler protocol: " + toString( scheduler->protocol ) + "\n";
 
@@ -637,7 +674,7 @@ string Daemon::dump_internals() const
     unsigned long idleLoad = 0;
     unsigned long niceLoad = 0;
 
-    if ( fill_stats( idleLoad, niceLoad, memory_fillgrade, &msg ) )
+    if ( fill_stats( idleLoad, niceLoad, memory_fillgrade, &msg, clients.active_processes ) )
     {
         result += "  cpu: " + toString( idleLoad ) + " idle, " +
                   toString( niceLoad ) + " nice\n";
@@ -679,32 +716,75 @@ int Daemon::scheduler_use_cs( UseCSMsg *msg )
     return 0;
 }
 
-bool Daemon::handle_transfer_env( MsgChannel *c, Msg *msg )
+bool Daemon::handle_transfer_env( MsgChannel *c, Msg *_msg )
 {
-    EnvTransferMsg *emsg = static_cast<EnvTransferMsg*>( msg );
+    log_error() << "handle_transfer_env" << endl;
+
+    Client *client = clients.find_by_channel( c );
+
+    assert(client);
+    assert(client->status != Client::TOINSTALL &&
+           client->status != Client::TOCOMPILE &&
+           client->status != Client::WAITCOMPILE);
+    assert(client->pipe_to_child < 0);
+
+    EnvTransferMsg *emsg = static_cast<EnvTransferMsg*>( _msg );
     string target = emsg->target;
     if ( target.empty() )
         target =  machine_name;
 
-    size_t installed_size = 0;
-    /* HACKALERT: install_environment can block for a while, make sure
-       the scheduler doesn't kick us */
-    if ( maybe_stats(true) )
-        installed_size = install_environment( envbasedir, emsg->target,
-                emsg->name, c, nobody_uid, nobody_gid );
-    else
-        return false;
-    if (!installed_size) {
-        trace() << "install environment failed" << endl;
-        c->send_msg(EndMsg()); // shut up, we had an error
-        return reannounce_environments();
+    int sock_to_stdin = -1;
+    FileChunkMsg* fmsg = 0;
+
+    pid_t pid = start_install_environment( envbasedir, emsg->target,
+        emsg->name, c, sock_to_stdin, fmsg, nobody_uid, nobody_gid );
+
+    if ( pid > 0) {
+        current_kids++;
+        client->status = Client::TOINSTALL;
+        client->outfile = emsg->target + "/" + emsg->name;
+        client->pipe_to_child = sock_to_stdin;
+        client->child_pid = pid;
+        handle_file_chunk_env(c, fmsg);
+        log_error() << "got pid " << pid << endl;
+    }
+    delete fmsg;
+    return pid > 0;
+}
+
+bool Daemon::handle_transfer_env_done( Client *client )
+{
+    log_error() << "handle_transfer_env_done" << endl;
+
+    assert(client);
+    assert(client->outfile.size());
+    assert(client->status == Client::TOINSTALL);
+
+    size_t installed_size = finalize_install_environment(envbasedir, client->outfile,
+                              client->child_pid, nobody_gid);
+
+    if (client->pipe_to_child >= 0) {
+        installed_size = 0;
+        close(client->pipe_to_child);
+        client->pipe_to_child = -1;
+    }
+
+    client->status = Client::UNKNOWN;
+    close(client->pipe_to_child);
+    client->pipe_to_child = -1;
+    string current = client->outfile;
+    client->outfile.clear();
+    client->child_pid = -1;
+    current_kids--;
+
+    log_error() << "installed_size: " << installed_size << endl;
+
+    if (installed_size) {
+        cache_size += installed_size;
+        envs_last_use[current] = time( NULL );
+        log_error() << "installed " << current << " size: " << installed_size
+            << " all: " << cache_size << endl;
     }
-    trace() << "envs " << dump_internals() << endl;
-    cache_size += installed_size;
-    string current = emsg->target + "/" + emsg->name;
-    envs_last_use[current] = time( NULL );
-    trace() << "installed " << emsg->name << " size: " << installed_size
-        << " all: " << cache_size << endl;
 
     time_t now = time( NULL );
     while ( cache_size > cache_size_limit ) {
@@ -719,6 +799,7 @@ bool Daemon::handle_transfer_env( MsgChannel *c, Msg *msg )
                 bool found = false;
                 for (Clients::const_iterator it2 = clients.begin(); it2 != clients.end(); ++it2)  {
                     if (it2->second->status == Client::TOCOMPILE ||
+                            it2->second->status == Client::TOINSTALL ||
                             it2->second->status == Client::WAITFORCHILD) {
 
                         assert( it2->second->job );
@@ -737,7 +818,7 @@ bool Daemon::handle_transfer_env( MsgChannel *c, Msg *msg )
         if ( oldest.empty() || oldest == current )
             break;
         size_t removed = remove_environment( envbasedir, oldest );
-        trace() << "removing " << envbasedir << " " << oldest << " " << oldest_time << " " << removed << endl;
+        trace() << "removing " << envbasedir << "/" << oldest << " " << oldest_time << " " << removed << endl;
         cache_size -= min( removed, cache_size );
         envs_last_use.erase( oldest );
     }
@@ -932,6 +1013,13 @@ void Daemon::handle_end( Client *client, int exitcode )
 #endif
     fd2chan.erase (client->channel->fd);
 
+    if (client->status == Client::TOINSTALL && client->pipe_to_child >= 0)
+    {
+        close(client->pipe_to_child);
+        client->pipe_to_child = -1;
+        handle_transfer_env_done(client);
+    }
+
     if ( client->status == Client::CLIENTWORK )
         clients.active_processes--;
 
@@ -940,6 +1028,17 @@ void Daemon::handle_end( Client *client, int exitcode )
  	client->job_id = 0;
     }
 
+    /* Delete from the clients map before send_scheduler, which causes a
+       double deletion. */
+    if (!clients.erase( client->channel ))
+    {
+        log_error() << "client can't be erased: " << client->channel << endl;
+        flush_debug();
+        log_error() << dump_internals() << endl;
+        flush_debug();
+        assert(false);
+    }
+
     if ( scheduler && client->status != Client::WAITFORCHILD ) {
         int job_id = client->job_id;
         if ( client->status == Client::TOCOMPILE )
@@ -960,6 +1059,7 @@ void Daemon::handle_end( Client *client, int exitcode )
             case Client::JOBDONE:
             case Client::WAITFORCHILD:
             case Client::LINKJOB:
+            case Client::TOINSTALL:
                 assert( false ); // should not have a job_id
                 break;
             case Client::WAITCOMPILE:
@@ -974,18 +1074,11 @@ void Daemon::handle_end( Client *client, int exitcode )
         } else if ( client->status == Client::CLIENTWORK ) {
             // Clientwork && !job_id == LINK
             trace() << "scheduler->send_msg( JobLocalDoneMsg( " << client->client_id << ") );\n";
-            send_scheduler( JobLocalDoneMsg( client->client_id ) );
+            if (!send_scheduler( JobLocalDoneMsg( client->client_id ) ))
+                trace() << "failed to reach scheduler for job done msg!" << endl;
         }
     }
 
-    if (!clients.erase( client->channel ))
-    {
-	log_error() << "client can't be erased: " << client->channel << endl;
-	flush_debug();
-	log_error() << dump_internals() << endl;
-	flush_debug();
-	assert(false);
-    }
     delete client;
 }
 
@@ -1053,6 +1146,55 @@ bool Daemon::handle_local_job( MsgChannel *c, Msg *msg )
     return true;
 }
 
+bool Daemon::handle_file_chunk_env(MsgChannel *c, Msg *msg)
+{
+    /* this sucks, we can block when we're writing
+       the file chunk to the child, but we can't let the child
+       handle MsgChannel itself due to MsgChannel's stupid
+       caching layer inbetween, which causes us to loose partial
+       data after the M_END msg of the env transfer.  */
+
+    Client *client = clients.find_by_channel( c );
+    assert (client);
+    assert (client->status == Client::TOINSTALL);
+
+    if (msg->type == M_FILE_CHUNK && client->pipe_to_child >= 0)
+    {
+        FileChunkMsg *fcmsg = static_cast<FileChunkMsg*>( msg );
+        ssize_t len = fcmsg->len;
+        off_t off = 0;
+        while ( len ) {
+            ssize_t bytes = write( client->pipe_to_child, fcmsg->buffer + off, len );
+            if ( bytes < 0 && errno == EINTR )
+                continue;
+
+            if ( bytes == -1 ) {
+                log_perror("write to transfer env pipe failed. ");
+
+                delete msg;
+                msg = 0;
+                handle_end(client, 137);
+                return false;
+            }
+
+            len -= bytes;
+            off += bytes;
+        }
+        return true;
+    }
+
+    if (msg->type == M_END) {
+        close(client->pipe_to_child);
+        client->pipe_to_child = -1;
+        return handle_transfer_env_done(client);
+    }
+
+    if (client->pipe_to_child >= 0)
+        handle_end(client, 138);
+
+    return false;
+}
+
 bool Daemon::handle_activity( MsgChannel *c )
 {
     Client *client = clients.find_by_channel( c );
@@ -1068,6 +1210,14 @@ bool Daemon::handle_activity( MsgChannel *c )
     }
 
     bool ret = false;
+    if (client->status == Client::TOINSTALL && client->pipe_to_child >= 0)
+        ret = handle_file_chunk_env(c, msg);
+
+    if (ret) {
+        delete msg;
+        return ret;
+    }
+
     switch ( msg->type ) {
     case M_GET_NATIVE_ENV: ret = handle_get_native_env( c ); break;
     case M_COMPILE_FILE: ret = handle_compile_file( c, msg ); break;
@@ -1130,7 +1280,8 @@ int Daemon::answer_client_requests()
             FD_SET (i, &listen_set);
         }
 
-        if (current_status == Client::WAITFORCHILD && client->pipe_to_child != -1) {
+        if (current_status == Client::WAITFORCHILD
+            && client->pipe_to_child != -1) {
             if (client->pipe_to_child > max_fd)
                 max_fd = client->pipe_to_child;
             FD_SET (client->pipe_to_child, &listen_set);
@@ -1160,19 +1311,21 @@ int Daemon::answer_client_requests()
     }
 
     if ( ret > 0 ) {
+        bool had_scheduler = scheduler;
         if ( scheduler && FD_ISSET( scheduler->fd, &listen_set ) ) {
             Msg *msg = scheduler->get_msg();
             if ( !msg ) {
                 log_error() << "scheduler closed connection\n";
                 close_scheduler();
+                clear_children();
                 return 1;
             } else {
                 ret = 0;
                 switch ( msg->type )
                 {
                 case M_PING:
-                    if ( !maybe_stats(true) )
-                        ret = 1;
+                    if (!IS_PROTOCOL_27(scheduler))
+                        ret = !send_scheduler(PingMsg());
                     break;
                 case M_USE_CS:
                     ret = scheduler_use_cs( dynamic_cast<UseCSMsg*>( msg ) );
@@ -1185,6 +1338,7 @@ int Daemon::answer_client_requests()
                     break;
                 default:
                     log_error() << "unknown scheduler type " << ( char )msg->type << endl;
+                    ret = 1;
                 }
             }
             delete msg;
@@ -1213,10 +1367,7 @@ int Daemon::answer_client_requests()
 
                 fd2chan[c->fd] = c;
                 c->read_a_bit();
-                while (c->has_msg()) {
-                    assert(client->status != Client::TOCOMPILE);
-                    if (!handle_activity (c))
-                        break;
+                while (c->has_msg() && handle_activity(c)) {
                     if (client->status == Client::TOCOMPILE ||
                             client->status == Client::WAITFORCHILD)
                         break;
@@ -1240,18 +1391,20 @@ int Daemon::answer_client_requests()
                 }
                 if (FD_ISSET (i, &listen_set)) {
                     c->read_a_bit();
-                    while (c->has_msg()) {
-                        assert(client->status != Client::TOCOMPILE);
-                        if (!handle_activity (c))
-                            break;
+                    while (c->has_msg() && handle_activity(c)) {
                         if (client->status == Client::TOCOMPILE ||
                                 client->status == Client::WAITFORCHILD)
                             break;
                     }
                     max_fd--;
                 }
-           }
+            }
         }
+        if ( had_scheduler && !scheduler ) {
+            clear_children();
+            return 2;
+        }
+
     }
     return 0;
 }
@@ -1287,17 +1440,12 @@ bool Daemon::reconnect()
     log_info() << "Connected to scheduler (" << remote_name << ")\n";
     current_load = -1000;
     gettimeofday( &last_stat, 0 );
-    last_scheduler_ping = last_stat.tv_sec;
     icecream_load = 0;
 
-    // perhaps our host name changed due to network change?
-    struct utsname uname_buf;
-    if ( !custom_nodename && !uname( &uname_buf ) )
-        nodename = uname_buf.nodename;
-
-    LoginMsg lmsg( PORT, nodename, machine_name );
+    LoginMsg lmsg( PORT, determine_nodename(), machine_name );
     lmsg.envs = available_environmnents(envbasedir);
     lmsg.max_kids = max_kids;
+    lmsg.noremote = noremote;
     return scheduler->send_msg( lmsg );
 }
 
@@ -1341,6 +1489,7 @@ int main( int argc, char ** argv )
             { "env-basedir", 1, NULL, 'b' },
             { "nobody-uid", 1, NULL, 'u'},
             { "cache-limit", 1, NULL, 0},
+            { "no-remote", 0, NULL, 0},
             { 0, 0, 0, 0 }
         };
 
@@ -1373,6 +1522,8 @@ int main( int argc, char ** argv )
                 }
                 else
                     usage("Error: --cache-limit requires argument");
+            } else if ( optname == "no-remote" ) {
+                d.noremote = true;
             }
 
         }
@@ -1462,18 +1613,7 @@ int main( int argc, char ** argv )
     log_info() << "ICECREAM daemon " VERSION " starting up (nice level "
                << nice_level << ") " << endl;
 
-    struct utsname uname_buf;
-    if ( uname( &uname_buf ) ) {
-        log_perror( "uname call failed" );
-        return 1;
-    }
-
-    if ( d.nodename.length() && d.nodename != uname_buf.nodename )
-        d.custom_nodename  = true;
-    if (!d.custom_nodename)
-        d.nodename = uname_buf.nodename;
-
-    d.machine_name = uname_buf.machine;
+    d.determine_system();
 
     chdir( "/" );
 
@@ -1514,6 +1654,14 @@ int main( int argc, char ** argv )
      * not.  */
     dcc_master_pid = getpid();
 
+    ofstream pidFile;
+    string progName = argv[0];
+    progName = progName.substr(progName.rfind('/')+1);
+    pidFilePath = string(RUNDIR)+string("/")+progName+string(".pid");
+    pidFile.open(pidFilePath.c_str());
+    pidFile << dcc_master_pid << endl;
+    pidFile.close();
+
     if ( !cleanup_cache( d.envbasedir ) )
         return 1;
 
