@@ -20,6 +20,7 @@
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
+#include <config.h>
 
 #include <sys/types.h>
 #include <netinet/in.h>
@@ -28,6 +29,10 @@
 #include <sys/select.h>
 #include <netinet/in.h>
 #include <netinet/tcp.h>
+#if HAVE_NETINET_TCP_VAR_H
+#include <sys/socketvar.h>
+#include <netinet/tcp_var.h>
+#endif
 #include <errno.h>
 #include <fcntl.h>
 #include <netdb.h>
@@ -38,7 +43,6 @@
 #include <assert.h>
 #include <minilzo.h>
 #include <stdio.h>
-#include <sys/utsname.h>
 
 #include "logging.h"
 #include "job.h"
@@ -159,12 +163,18 @@ MsgChannel::update_state (void)
     case NEED_LEN:
       if (text_based)
         {
-	  /* XXX handle other line endings.  */
-	  void *lineend = memchr (inbuf + intogo, '\r', inofs - intogo);
-	  if (!lineend)
-	    break;
-	  instate = HAS_MSG;
-	  goto has_msg;
+          // Skip any leading whitespace
+          for (;inofs < intogo; ++inofs)
+             if (inbuf[inofs] >= ' ') break;
+
+          // Skip until next newline
+          for(inmsglen = 0; inmsglen < inofs - intogo; ++inmsglen)
+             if (inbuf[intogo+inmsglen] < ' ')
+               {
+                 instate = HAS_MSG;
+                 break;
+               }
+          break;
 	}
       else if (inofs - intogo >= 4)
         {
@@ -186,7 +196,6 @@ MsgChannel::update_state (void)
       else
         break;
     case HAS_MSG:
-    has_msg:
       /* handled elsewere */
       break;
     }
@@ -240,7 +249,15 @@ MsgChannel::flush_writebuf (bool blocking)
   bool error = false;
   while (msgtogo)
     {
+#ifdef MSG_NOSIGNAL
       ssize_t ret = send (fd, buf, msgtogo, MSG_NOSIGNAL);
+#else
+      void (*oldsigpipe)(int);
+
+      oldsigpipe = signal (SIGPIPE, SIG_IGN);
+      ssize_t ret = send (fd, buf, msgtogo, 0);
+      signal (SIGPIPE, oldsigpipe);
+#endif
       if (ret < 0)
         {
 	  if (errno == EINTR)
@@ -290,7 +307,7 @@ MsgChannel::readuint32 (uint32_t &buf)
 {
   if (inofs >= intogo + 4)
     {
-      if ( off_t(inbuf + intogo) % 4 ) {
+      if ( ptrdiff_t(inbuf + intogo) % 4 ) {
         char t_buf[4];
         memcpy(t_buf, inbuf + intogo, 4);
         buf = *(uint32_t *)t_buf;
@@ -422,7 +439,8 @@ MsgChannel::readcompressed (unsigned char **uncompressed_buf,
 	     Remove the buffer, and indicate there is nothing in it,
 	     but don't reset the compressed_len, so our caller know,
 	     that there actually was something read in.  */
-          log_error() << "internal error - decompression failed: " << ret << endl;
+          log_error() << "internal error - decompression of data from " << dump().c_str()
+              << " failed: " << ret << endl;
 	  delete [] *uncompressed_buf;
 	  *uncompressed_buf = 0;
 	  uncompressed_len = 0;
@@ -478,18 +496,10 @@ MsgChannel::read_line (string &line)
     }
   else
     {
-      const char *linebeg = inbuf + intogo;
-      const char *lineend = (const char *) memchr (linebeg, '\r', inofs - intogo);
-      if (!lineend)
-        line = "";
-      else
-        {
-	  /* Without the EOL */
-	  line = string(linebeg, lineend - linebeg);
-	  intogo += lineend - linebeg + 1;
-	  if (intogo < inofs && inbuf[intogo] == '\n')
-	    intogo++;
-	}
+	  line = string(inbuf + intogo, inmsglen);
+	  intogo += inmsglen;
+          while (intogo < inofs && inbuf[intogo] < ' ')
+            intogo++;
     }
 }
 
@@ -593,7 +603,7 @@ MsgChannel *Service::createChannel (const string &hostname, unsigned short p, in
   remote_addr.sin_family = AF_INET;
   remote_addr.sin_port = htons (p);
   memcpy (&remote_addr.sin_addr.s_addr, host->h_addr_list[0], host->h_length);
-  setsockopt(remote_fd, IPPROTO_TCP, TCP_NODELAY, (char*) &i, sizeof(i));
+  setsockopt (remote_fd, IPPROTO_TCP, TCP_NODELAY, (char*) &i, sizeof(i));
 
   if ( timeout )
     {
@@ -673,6 +683,34 @@ MsgChannel::MsgChannel (int _fd, struct sockaddr *_a, socklen_t _l, bool text)
   eof = false;
   text_based = text;
 
+  int on = 1;
+  if (!setsockopt (_fd, SOL_SOCKET, SO_KEEPALIVE, (char*) &on, sizeof(on)))
+  {
+#if defined( TCP_KEEPIDLE )
+      int keepidle = TCP_KEEPIDLE;
+#else
+      int keepidle = TCPCTL_KEEPIDLE;
+#endif
+
+      int sec;
+      sec = MAX_SCHEDULER_PING - 3 * MAX_SCHEDULER_PONG;
+      setsockopt (_fd, IPPROTO_TCP, keepidle, (char*) &sec, sizeof(sec));
+
+#if defined( TCP_KEEPINTVL )
+      int keepintvl = TCP_KEEPINTVL;
+#else
+      int keepintvl = TCPCTL_KEEPINTVL;
+#endif
+
+      sec = MAX_SCHEDULER_PONG;
+      setsockopt (_fd, IPPROTO_TCP, keepintvl, (char*) &sec, sizeof(sec));
+
+#ifdef TCP_KEEPCNT
+      sec = 3;
+      setsockopt (_fd, IPPROTO_TCP, TCP_KEEPCNT, (char*) &sec, sizeof(sec));
+#endif
+  }
+
   if (fcntl (fd, F_SETFL, O_NONBLOCK) < 0)
     log_perror("MsgChannel fcntl()");
 
@@ -843,8 +881,10 @@ MsgChannel::get_msg(int timeout)
     case M_CS_CONF: m = new ConfCSMsg; break;
     case M_TIMEOUT: break;
     }
-  if (!m)
+  if (!m) {
+      trace() << "no message type" << endl;
     return 0;
+  }
   m->fill_from_channel (this);
   instate = NEED_LEN;
   update_state ();
@@ -1335,7 +1375,7 @@ JobDoneMsg::send_to_channel (MsgChannel *c) const
 }
 
 LoginMsg::LoginMsg(unsigned int myport, const std::string &_nodename, const std::string _host_platform)
-  : Msg(M_LOGIN), port( myport ), chroot_possible( false ), nodename( _nodename ), host_platform( _host_platform )
+  : Msg(M_LOGIN), port( myport ), noremote( false ), chroot_possible( false ), nodename( _nodename ), host_platform( _host_platform )
 {
   // check if we're root
   chroot_possible = ( geteuid() == 0 );
@@ -1353,6 +1393,9 @@ LoginMsg::fill_from_channel (MsgChannel *c)
   uint32_t net_chroot_possible = 0;
   c->readuint32 (net_chroot_possible);
   chroot_possible = net_chroot_possible != 0;
+  uint32_t net_noremote = 0;
+  if (IS_PROTOCOL_26( c )) c->readuint32 (net_noremote);
+  noremote = (net_noremote != 0);
 }
 
 void
@@ -1365,6 +1408,7 @@ LoginMsg::send_to_channel (MsgChannel *c) const
   c->write_string( nodename );
   c->write_string( host_platform );
   c->writeuint32( chroot_possible );
+  if (IS_PROTOCOL_26( c )) c->writeuint32( noremote );
 }
 
 void
