@@ -42,11 +42,14 @@
 #include <queue>
 #include <algorithm>
 #include <cassert>
+#include <fstream>
+#include <string>
 #include <stdio.h>
 #include "comm.h"
 #include "logging.h"
 #include "job.h"
 #include "config.h"
+#include "bench.h"
 
 #define DEBUG_SCHEDULER 0
 
@@ -82,6 +85,8 @@
 
 using namespace std;
 
+static string pidFilePath;
+
 struct JobStat {
   unsigned long osize;  // output size (uncompressed)
   unsigned long compile_time_real;  // in milliseconds
@@ -141,11 +146,13 @@ public:
   // LOAD is load * 1000
   unsigned int load;
   int max_jobs;
+  bool noremote;
   list<Job*> joblist;
   Environments compiler_versions;  // Available compilers
   CS (int fd, struct sockaddr *_addr, socklen_t _len, bool text_based)
-    : MsgChannel(fd, _addr, _len, text_based), load(1000), max_jobs(0), state(CONNECTED),
-      type(UNKNOWN), chroot_possible(false)
+    : MsgChannel(fd, _addr, _len, text_based),
+      load(1000), max_jobs(0), noremote(false),
+      state(CONNECTED), type(UNKNOWN), chroot_possible(false)
   {
     hostid = 0;
     busy_installing = 0;
@@ -164,12 +171,14 @@ public:
   bool chroot_possible;
   static unsigned int hostid_counter;
   map<int, int> client_map; // map client ID for daemon to our IDs
+  bool is_eligible( const Job *job );
+  bool check_remote( const Job *job ) const;
 };
 
 unsigned int CS::hostid_counter = 0;
 
 static map<int, MsgChannel *> fd2chan;
-static bool allow_run_as_user = false;
+static bool exit_main_loop = false;
 
 time_t starttime;
 
@@ -390,6 +399,8 @@ handle_monitor_stats( CS *cs, StatsMsg *m = 0)
   msg += buffer;
   sprintf( buffer, "MaxJobs:%d\n", cs->max_jobs );
   msg += buffer;
+  sprintf( buffer, "NoRemote:%s\n", cs->noremote ? "true" : "false"  );
+  msg += buffer;
   sprintf( buffer, "Platform:%s\n", cs->host_platform.c_str() );
   msg += buffer;
   sprintf( buffer, "Speed:%f\n", server_speed( cs ) );
@@ -647,6 +658,23 @@ can_install( CS* cs, const Job *job )
   return string();
 }
 
+bool CS::check_remote( const Job *job ) const
+{
+    bool local = (job->submitter == this);
+    return local || !noremote;
+}
+
+bool CS::is_eligible( const Job *job )
+{
+  bool jobs_okay = int( joblist.size() ) < max_jobs;
+  bool load_okay = load < 1000;
+  return jobs_okay
+    && chroot_possible
+    && load_okay
+    && can_install( this, job ).size()
+    && this->check_remote( job );
+}
+
 static CS *
 pick_server(Job *job)
 {
@@ -685,9 +713,7 @@ pick_server(Job *job)
 	for (it = css.begin(); it != css.end(); ++it)
           {
 	     if ((*it)->nodename == job->preferred_host
-                 && int( (*it)->joblist.size() ) < (*it)->max_jobs
-                 && (*it)->chroot_possible
-                 && (*it)->load < 1000 && can_install( *it, job ).size() )
+                 && (*it)->is_eligible( job ))
 	       return *it;
 	  }
         return 0;
@@ -699,9 +725,7 @@ pick_server(Job *job)
       for (it = css.begin(); it != css.end(); ++it)
         {
           trace() << "no job stats - looking at " << ( *it )->nodename << " load: " << (*it )->load << " can install: " << can_install( *it, job ) << endl;
-          if (int( (*it)->joblist.size() ) < (*it)->max_jobs
-	      && (*it)->chroot_possible
-              && (*it)->load < 1000 && can_install( *it, job ).size() )
+          if ((*it)->is_eligible( job ))
             {
               trace() << "returning first " << ( *it )->nodename << endl;
               return *it;
@@ -726,6 +750,8 @@ pick_server(Job *job)
   CS *best = 0;
   // best uninstalled
   CS *bestui = 0;
+  // best preloadable host
+  CS *bestpre = 0;
 
   uint matches = 0;
 
@@ -733,7 +759,8 @@ pick_server(Job *job)
     {
       CS *cs = *it;
       /* For now ignore overloaded servers.  */
-      if (int( cs->joblist.size() ) >= cs->max_jobs || cs->load >= 1000)
+      /* Pre-loadable (cs->joblist.size()) == (cs->max_jobs) is checked later.  */
+      if (int( cs->joblist.size() ) > cs->max_jobs || cs->load >= 1000)
         {
 #if DEBUG_SCHEDULER > 1
           trace() << "overloaded " << cs->nodename << " " << cs->joblist.size() << "/" <<  cs->max_jobs << " jobs, load:" << cs->load << endl;
@@ -761,12 +788,19 @@ pick_server(Job *job)
         }
 
       /* Don't use non-chroot-able daemons for remote jobs.  XXX */
-      if (!allow_run_as_user && !cs->chroot_possible)
+      if (!cs->chroot_possible)
         {
 	  trace() << cs->nodename << " can't use chroot\n";
 	  continue;
 	}
 
+      // Check if remote & if remote allowed
+      if (!cs->check_remote( job ))
+        {
+	  trace() << cs->nodename << " fails remote job check\n";
+	  continue;
+	}
+
 
 #if DEBUG_SCHEDULER > 1
       trace() << cs->nodename << " compiled " << cs->last_compiled_jobs.size() << " got now: " <<
@@ -801,7 +835,12 @@ pick_server(Job *job)
           else
             if (best->last_compiled_jobs.size() != 0
                 && server_speed (best, job) < server_speed (cs, job))
-              best = cs;
+              {
+                if (int( cs->joblist.size() ) < cs->max_jobs)
+                  best = cs;
+                else
+                  bestpre = cs;
+              }
           matches++;
         }
       else
@@ -813,7 +852,12 @@ pick_server(Job *job)
           else
             if (bestui->last_compiled_jobs.size() != 0
                 && server_speed (bestui, job) < server_speed (cs, job))
-              bestui = cs;
+              {
+                if (int( cs->joblist.size() ) < cs->max_jobs)
+                  bestui = cs;
+                else
+                  bestpre = cs;
+              }
         }
     }
 
@@ -830,11 +874,22 @@ pick_server(Job *job)
       return best;
     }
 
-#if DEBUG_SCHEDULER > 1
   if ( bestui )
-    trace() << "taking best uninstalled " << bestui->nodename << " " <<  server_speed (bestui) << endl;
+    {
+#if DEBUG_SCHEDULER > 1
+      trace() << "taking best uninstalled " << bestui->nodename << " " <<  server_speed (bestui) << endl;
 #endif
-  return bestui;
+      return bestui;
+    }
+
+  if ( bestpre )
+    {
+#if DEBUG_SCHEDULER > 1
+      trace() << "taking best preload " << bestui->nodename << " " <<  server_speed (bestui) << endl;
+#endif
+    }
+
+  return bestpre;
 }
 
 /* Prunes the list of connected servers by those which haven't
@@ -850,13 +905,17 @@ prune_servers ()
 
   for (it = css.begin(); it != css.end(); )
     {
-      if ( now - ( *it )->last_talk >= MAX_SCHEDULER_PING )
+      /* protocol version 27 and newer use TCP keepalive */
+      if (IS_PROTOCOL_27(*it)) {
+        ++it;
+        continue;
+      }
+
+      if (now - ( *it )->last_talk >= MAX_SCHEDULER_PING)
         {
-          if ( ( *it )->max_jobs >= 0 )
+          if (( *it )->max_jobs >= 0)
             {
-#if DEBUG_SCHEDULER > 1
               trace() << "send ping " << ( *it )->nodename << endl;
-#endif
               ( *it )->max_jobs *= -1; // better not give it away
               if(( *it )->send_msg( PingMsg() ))
 		{
@@ -1038,9 +1097,6 @@ empty_queue()
   return true;
 }
 
-// unused
-const char icecream_bench_code[] = "\n";
-
 static bool
 handle_login (MsgChannel *c, Msg *_m)
 {
@@ -1050,7 +1106,7 @@ handle_login (MsgChannel *c, Msg *_m)
 
   /* If we don't allow non-chroot-able daemons in the farm,
      discard them here.  */
-  if (!allow_run_as_user && !m->chroot_possible)
+  if (!m->chroot_possible)
     return false;
 
   std::ostream& dbg = trace();
@@ -1060,6 +1116,7 @@ handle_login (MsgChannel *c, Msg *_m)
   cs->remote_port = m->port;
   cs->compiler_versions = m->envs;
   cs->max_jobs = m->max_kids;
+  cs->noremote = m->noremote;
   if ( m->nodename.length() )
     cs->nodename = m->nodename;
   else
@@ -1068,6 +1125,20 @@ handle_login (MsgChannel *c, Msg *_m)
   cs->chroot_possible = m->chroot_possible;
   cs->pick_new_id();
   handle_monitor_stats( cs );
+
+  /* remove any other clients with the same IP, they must be stale */
+  for (list<CS*>::iterator it = css.begin(); it != css.end(); )
+    {
+      if (cs->eq_ip(*(*it)))
+      {
+        CS* old = *it;
+        ++it;
+        handle_end(old, 0);
+        continue;
+      }
+      ++it;
+    }
+
   css.push_back (cs);
 
 #if 1
@@ -1107,7 +1178,7 @@ handle_relogin (MsgChannel *c, Msg *_m)
   if (IS_PROTOCOL_24( c ))
     c->send_msg (ConfCSMsg(icecream_bench_code));
 
-  return true;
+  return false;
 }
 
 static bool
@@ -1174,7 +1245,7 @@ handle_job_done (MsgChannel *c, Msg *_m)
       for (mit = jobs.begin(); mit != jobs.end(); mit++)
         {
           Job *job = mit->second;
-          trace() << "looking for waitcs " << job->server << " " << job->submitter  << " " << c << " " 
+          trace() << "looking for waitcs " << job->server << " " << job->submitter  << " " << c << " "
 		  << job->state << " " << job->local_client_id << " " << m->job_id << endl;
           if (job->server == 0 && job->submitter == c && job->state == Job::PENDING
               && job->local_client_id == m->job_id )
@@ -1193,7 +1264,7 @@ handle_job_done (MsgChannel *c, Msg *_m)
 		    list<Job*>::iterator jit;
 		    for (jit = l->l.begin(); jit != l->l.end(); ++jit)
 		      {
-			if (*jit == j) 
+			if (*jit == j)
 			  {
 			    l->l.erase(jit);
 			    break;
@@ -1218,6 +1289,33 @@ handle_job_done (MsgChannel *c, Msg *_m)
       return false;
     }
 
+  if (j->state == Job::PENDING)
+    {
+      trace() << "job ID still pending ?! scheduler recently restarted? " << m->job_id << endl;
+      return false;
+    }
+
+  if (m->is_from_server() && j->server != c)
+    {
+      log_info() << "the server isn't the same for job " << m->job_id << endl;
+      log_info() << "server: " << j->server->nodename << endl;
+      log_info() << "msg came from: " << ((CS*)c)->nodename << endl;
+      // the daemon is not following matz's rules: kick him
+      handle_end(c, 0);
+      return false;
+    }
+  if (!m->is_from_server() && j->submitter != c)
+    {
+      log_info() << "the submitter isn't the same for job " << m->job_id << endl;
+      log_info() << "submitter: " << j->submitter->nodename << endl;
+      log_info() << "msg came from: " << ((CS*)c)->nodename << endl;
+      // the daemon is not following matz's rules: kick him
+      handle_end(c, 0);
+      return false;
+    }
+
+
+
   if ( m->exitcode == 0 )
     {
       std::ostream &dbg = trace();
@@ -1247,26 +1345,7 @@ handle_job_done (MsgChannel *c, Msg *_m)
     trace() << "END " << m->job_id
 	    << " status=" << m->exitcode << endl;
 
-  if (m->is_from_server() && j->server != c)
-    {
-      log_info() << "the server isn't the same for job " << m->job_id << endl;
-      log_info() << "server: " << j->server->nodename << endl;
-      log_info() << "msg came from: " << ((CS*)c)->nodename << endl;
-      // the daemon is not following matz's rules: kick him
-      handle_end(c, 0);
-      return false;
-    }
-  if (!m->is_from_server() && j->submitter != c)
-    {
-      log_info() << "the submitter isn't the same for job " << m->job_id << endl;
-      log_info() << "submitter: " << j->submitter->nodename << endl;
-      log_info() << "msg came from: " << ((CS*)c)->nodename << endl;
-      // the daemon is not following matz's rules: kick him
-      handle_end(c, 0);
-      return false;
-    }
-
-  if (j->server) 
+  if (j->server)
     {
       j->server->joblist.remove (j);
       j->server->busy_installing = 0;
@@ -1726,13 +1805,27 @@ usage(const char* reason = 0)
        << "  -h, --help\n"
        << "  -l, --log-file <file>\n"
        << "  -d, --daemonize\n"
-       << "  -r, --allow-run-as-user\n"
        << "  -v[v[v]]]\n"
        << endl;
 
   exit(1);
 }
 
+static void
+trigger_exit( int signum )
+{
+  if (!exit_main_loop)
+    exit_main_loop = true;
+  else
+    {
+      // hmm, we got killed already. try better
+      cerr << "forced exit." << endl;
+      _exit(1);
+    }
+  // make BSD happy
+  signal(signum, trigger_exit);
+}
+
 int
 main (int argc, char * argv[])
 {
@@ -1753,7 +1846,6 @@ main (int argc, char * argv[])
       { "port", 0, NULL, 'p' },
       { "daemonize", 0, NULL, 'd'},
       { "log-file", 1, NULL, 'l'},
-      { "allow-run-as-user", 1, NULL, 'u' },
       { 0, 0, 0, 0 }
     };
 
@@ -1801,10 +1893,6 @@ main (int argc, char * argv[])
         usage("Error: -p requires argument");
       break;
 
-    case 'r':
-      allow_run_as_user = true;
-      break;
-
     default:
       usage();
     }
@@ -1835,7 +1923,19 @@ main (int argc, char * argv[])
 
   starttime = time( 0 );
 
-  while (1)
+  ofstream pidFile;
+  string progName = argv[0];
+  progName = progName.substr(progName.rfind('/')+1);
+  pidFilePath = string(RUNDIR)+string("/")+progName+string(".pid");
+  pidFile.open(pidFilePath.c_str());
+  pidFile << getpid() << endl;
+  pidFile.close();
+
+  signal(SIGTERM, trigger_exit);
+  signal(SIGINT, trigger_exit);
+  signal(SIGALRM, trigger_exit);
+
+  while (!exit_main_loop)
     {
       struct timeval tv;
       tv.tv_usec = 0;
@@ -1909,8 +2009,9 @@ main (int argc, char * argv[])
                   continue;
                 }
 	      fd2chan[cs->fd] = cs;
-	      if (!cs->read_a_bit () || cs->has_msg ())
-	        handle_activity (cs);
+	      while (!cs->read_a_bit () || cs->has_msg ())
+                if(! handle_activity (cs))
+                  break;
 	    }
         }
       if (max_fd && FD_ISSET (text_fd, &read_set))
@@ -1936,8 +2037,9 @@ main (int argc, char * argv[])
                   continue;
                 }
 	      fd2chan[cs->fd] = cs;
-	      if (!cs->read_a_bit () || cs->has_msg ())
-	        handle_activity (cs);
+	      while (!cs->read_a_bit () || cs->has_msg ())
+	        if (!handle_activity (cs))
+                  break;
 	    }
 	}
       if (max_fd && FD_ISSET (broad_fd, &read_set))
@@ -1959,7 +2061,8 @@ main (int argc, char * argv[])
 	      if (err != EAGAIN)
 	        return -1;
 	    }
-	  else
+            /* Only answer if daemon would be able to talk to us. */
+	  else if (buf[0] >= MIN_PROTOCOL_VERSION)
 	    {
 	      log_info() << "broadcast from " << inet_ntoa (broad_addr.sin_addr)
                          << ":" << ntohs (broad_addr.sin_port) << "\n";
@@ -1985,13 +2088,16 @@ main (int argc, char * argv[])
           ++it;
           if (FD_ISSET (i, &read_set))
             {
-              if (!c->read_a_bit () || c->has_msg ())
-                handle_activity (c);
+              while (!c->read_a_bit () || c->has_msg ())
+                if(!handle_activity (c))
+                  break;
               max_fd--;
             }
         }
     }
   close (broad_fd);
+  shutdown (broad_fd, SHUT_RDWR);
+  unlink(pidFilePath.c_str());
   return 0;
 }
 /*
